#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

import json
import decimal
from datetime import datetime, date
from collections import Mapping, Iterable
from urlparse import urljoin
from flask import Flask, request
from exceptions import RestHttpError


class CustomJsonEncoder(json.JSONEncoder):

    DATETIME_FMT = "%Y-%m-%d %H:%M:%S"

    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.strftime(self.DATETIME_FMT)
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, Mapping):
            return dict(obj)
        elif isinstance(obj, Iterable):
            return list(obj)

        to_dict = getattr(
            obj,
            '_as_dict',
            None) or getattr(
            obj,
            'to_dict',
            None)
        if to_dict and callable(to_dict):
            return to_dict()

        return super(CustomJsonEncoder, self).default(obj)


def flask_res(data=None, code=200):
    data = dict(msg='ok', code=200) if not data else data
    return Flask.response_class(json.dumps(
        data, cls=CustomJsonEncoder), status=code, mimetype='application/json')


def rest_abort(code):
    raise RestHttpError(code)


def get_bp_prefix(app, bp_name):
    prefix_pattern = app.config.get('BP_PREFIX_PATTERN', '/api/{0}/{1}/')
    default_prefix = prefix_pattern.format(app.config['VERSION'], bp_name)
    if not app.config.get('BP_PREFIX', None):
        app.config['BP_PREFIX'] = dict()
    app.config['BP_PREFIX'].setdefault(bp_name, default_prefix)
    return app.config['BP_PREFIX'][bp_name]


def get_relation_url(table_name, pk, relation_name):
    from main import app
    bp_prefix = get_bp_prefix(app, 'rest_db')
    relation_path = '/'.join((table_name, str(pk), relation_name))
    return urljoin(urljoin(request.host_url, bp_prefix), relation_path)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from copy import deepcopy
from functools import wraps
from flask import request, Flask, Response, current_app
from app.lib.utils import flask_res
from app.lib.validator import Validator


class RestRoute(object):

    def __init__(self, bp, rule="", methods=None, validator=None,
                 pre_processes=None):
        self.rule = rule
        self.methods = methods or ['GET']
        self.validator = Validator(validator) if validator else None
        self.pre_processes = pre_processes or []
        self.bp = bp

    def __call__(self, view_function, endpoint=None):
        view_endpoint = endpoint or view_function.__name__

        @self.bp.route(rule=self.rule, methods=self.methods,
                       endpoint=view_endpoint)
        @wraps(view_function)
        def wrapped(**kwargs):
            data = deepcopy(kwargs)
            http_data = self.get_http_data()
            data.update(http_data)
            self.pre_process(data)
            self.result = view_function(data)
            return self.post_process(self.result)

        return wrapped

    def pre_process(self, data):
        if self.validator:
            self.validator(data)

        for process in self.pre_processes:
            process(data)

    @staticmethod
    def post_process(result):
        if isinstance(result, (Response, Flask.response_class)):
            # 用户已经封装好返回值
            return result
        else:
            return flask_res({
                'code': 200,
                'msg': 'ok',
                'result': result
            })

    @staticmethod
    def get_http_data():
        http_data = dict()
        http_data.update(request.args.to_dict())
        try:
            http_data.update(request.json)
        except Exception as e:
            current_app.logger.info(u'no json data {0}'.format(str(e)))

        return http_data


def rest_route(bp, rule="", methods=None, validator=None, pre_processes=None):
    return RestRoute(bp, rule, methods, validator, pre_processes)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from app.lib.pine_wrapper import PineFlask, PineBlueprint
from app import init_app

app = PineFlask(__name__, instance_relative_config=True)
api = PineBlueprint('api', __name__)
app = init_app(app=app, bp_list=[api])



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/6 by allen
"""

import weakref
from passlib.context import LazyCryptContext
from sqlalchemy import types
from sqlalchemy.ext.mutable import Mutable
from app.lib.database.utils import ConversionBase


class Password(Mutable, object):

    @staticmethod
    def coerce(cls, key, value):
        if isinstance(value, Password):
            return value

        if isinstance(value, basestring):
            return cls(value, secret=True)

        super(Password, cls).coerce(key, value)

    def __init__(self, value, context, secret=False):
        self.hash = value if not secret else None
        self.secret = value if secret else None

        if isinstance(self.hash, unicode):
            self.hash = self.hash.encode("utf-8")

        self.context = weakref.proxy(context) if context is not None else None

    def __eq__(self, value):
        if self.hash is None or value is None:
            return self.hash is None

        if isinstance(value, Password):
            return self.hash == value.hash

        if self.context is None:
            return value == self

        if isinstance(value, basestring):
            valid, new = self.context.verify_and_update(value, self.hash)

            if valid and new:
                self.hash = new

            if isinstance(self.hash, unicode):
                self.hash = self.hash.encode('utf-8')
                self.changed()

            return valid

        return False

    def _as_dict(self):
        return self.hash


class PasswordType(types.TypeDecorator, ConversionBase):

    python_type = Password
    impl = types.VARBINARY(1024)

    def __init__(self, max_length=None, **kwargs):
        self._max_length = max_length
        self.context = LazyCryptContext(**kwargs)

        super(PasswordType, self).__init__()

    @property
    def length(self):
        if self._max_length is None:
            self._max_length = self.calculate_max_length()

        return self._max_length

    def calculate_max_length(self):

        max_lengths = [1024]
        for name in self.context.schemes():
            scheme = getattr(__import__('passlib.hash').hash, name)
            length = 4 + len(scheme.name)
            length += len(str(getattr(scheme, 'max_rounds', '')))
            length += (getattr(scheme, 'max_salt_size', 0) or 0)
            length += getattr(
                scheme,
                'encoded_checksum_size',
                scheme.checksum_size
            )
            max_lengths.append(length)

        return max(max_lengths)

    def load_dialect_impl(self, dialect):
        impl = types.VARBINARY(self.length)
        return dialect.type_descriptor(impl)

    def process_bind_param(self, value, dialect):
        if isinstance(value, Password):
            if value.secret is not None:
                return value.context.encrypt(value.secret).encode("utf-8")

            return value.hash

        if isinstance(value, basestring):
            return self.context.encrypt(value).encode("utf-8")

    def process_result_value(self, value, dialect):
        if value is not None:
            return Password(value, self.context)

    def _conversion(self, value):
        if value is None:
            return

        if not isinstance(value, Password):
            value = self.context.encrypt(value).encode("utf-8")
            return Password(value, context=self.context, secret=False)

        else:
            value.context = weakref.proxy(self.context)
            if value.secret:
                value.hash = value.context.encrypt(
                    value.secret).encode("utf-8")
                value.secret = None

            return value



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/2 by allen
"""

import codecs
import json
import os
import shutil
import time
from copy import deepcopy
from subprocess import check_output

from tadpole import config
from tadpole import logger


class _PyConfigMixin(object):
    def __init__(self):
        pass

    @staticmethod
    def process_custom_config(config_dict, custom_processes):
        config_lines, import_lines = list(), list()

        if custom_processes:
            for key in set(custom_processes).intersection(
                    set(config_dict.keys())):
                custom_imports, custom_config = custom_processes[key](
                    key, config_dict.pop(key))
                import_lines.extend(custom_imports)
                config_lines.extend(custom_config)

        return config_dict, import_lines, config_lines

    @staticmethod
    def replace_fmt(config_content, fmt_dict):
        for key in fmt_dict:
            fmt_key = "{%s}" % key
            if fmt_key in config_content:
                config_content = config_content.replace(fmt_key, fmt_dict[key])

        return config_content

    @staticmethod
    def process_key_value(key, value):
        if isinstance(value, unicode):
            return u"{key} = u'{value}'\n".format(key=key, value=value)
        if isinstance(value, basestring):
            return u"{key} = '{value}'\n".format(key=key, value=value)
        elif isinstance(value, (list, dict)):
            return u"{key} = {value}\n".format(
                key=key, value=json.dumps(value, indent=4, ensure_ascii=False))

        return u"{key} = {value}\n".format(key=key, value=value)

    def render_config(self, config_dict, custom_processes=None, fmt_dict=None):

        config_dict = deepcopy(config_dict)
        config_dict, import_lines, config_lines = self.process_custom_config(
            config_dict, custom_processes)

        keys = config_dict['__KEYS_ORDER'] if config_dict.get(
            '__KEYS_ORDER') else config_dict.keys()

        for key in keys:
            if key == '\n':
                config_lines.append(key)
                continue
            elif key not in config_dict or key.startswith('__'):
                continue
            line = self.process_key_value(key, config_dict[key])
            config_lines.append(line)

        config_content = u"{headers}\n{import_lines}\n{config_lines}".format(
            headers=config.TEMPLATE_HEADER, import_lines=u''.join(
                import_lines),
            config_lines=u''.join(config_lines))

        if fmt_dict:
            config_content = self.replace_fmt(config_content, fmt_dict)

        return config_content


class _InitHandler(_PyConfigMixin):

    def __init__(self, project_name, version, owner, email):

        super(_InitHandler, self).__init__()

        self.project_name = project_name
        self.version = version
        self.owner = owner
        self.email = email
        self.work_dir = os.path.join(os.getcwd(), self.project_name)
        self.cur_dir = os.path.dirname(os.path.realpath(__file__))
        self.template_dir = os.path.join(self.cur_dir, config.TEMPLATE_SRC)
        self.fmt_dict = {
            'app_name': self.project_name.lower(),
            'version': self.version.lower(),
            'owner': self.owner.lower(),
            'email': self.email
        }

    @staticmethod
    def _make_dirs():
        directories = config.TEMPLATE_DIRS_CREATE
        logger.debug(u"make dirs {0}".format(u','.join(directories)))

        for directory in directories:
            if not os.path.exists(directory):
                os.mkdir(directory)

    @staticmethod
    def _process_gun_workers(key, workers):
        import_lines, config_lines = list(), list()

        if isinstance(workers, int) or (
                isinstance(workers, basestring) and workers.isdigit()):
            config_lines.append('{0} = {1}\n'.format(key, workers))
        else:
            try:
                wp_cpu = int(workers.split("*")[0])
            except Exception as e:
                logger.debug(str(e))
                wp_cpu = 2

            if wp_cpu < 1 or wp_cpu > 10:
                wp_cpu = 2

            import_lines.append("import multiprocessing\n")
            config_lines.append(
                "{0} = multiprocessing.cpu_count() * {1} \n".format(key, wp_cpu))

        return import_lines, config_lines

    def _get_project_config(self):
        config_dict = deepcopy(config.TEMPLATE_PROJECT)
        config_dict.update({
            'APP_NAME': self.project_name,
            'VERSION': self.version,
            'OWNER': self.owner,
            'EMAIL': self.email
        })

        return config_dict

    def _init_config(self):

        # init config.py
        logger.debug("make config {0}".format(config.TEMPLATE_CONF_NAME))
        config_dict = self._get_project_config()
        config_content = self.render_config(
            config_dict, fmt_dict=self.fmt_dict)
        with codecs.open(config.TEMPLATE_CONF_NAME, "w", "utf-8") as cfp:
            cfp.write(config_content)

        # init instance/config.py
        instance_config = "instance/{0}".format(config.TEMPLATE_CONF_NAME)
        logger.debug("make instance config {0}".format(instance_config))
        if "instance/" in config.TEMPLATE_DIRS_CREATE:
            instance_dict = config.TEMPLATE_INSTANCE
            config_content = self.render_config(instance_dict)
            with codecs.open(instance_config, "w", "utf-8") as ifp:
                ifp.write(config_content)

        # init gun.py
        logger.debug("make config {0}".format(config.TEMPLATE_GUN_CONF))
        config_content = self.render_config(config.TEMPLATE_GUNICORN, {
            'workers': self._process_gun_workers})
        with codecs.open(config.TEMPLATE_GUN_CONF, "w", "utf-8") as gfp:
            gfp.write(config_content)

    def _init_template(self):
        logger.debug(
            "gen %s file for start, stop ... actions" %
            self.project_name)
        with codecs.open("template.py", "r", "utf-8") as tfp:
            content = tfp.read()

        with codecs.open(self.project_name, "w", "utf-8") as pfp:
            content = content.replace("{{template}}", self.project_name)
            pfp.write(content)

        os.chmod(self.project_name, 0o740)
        os.remove("template.py")

    @staticmethod
    def _init_venv():
        logger.debug("make venv")
        out = check_output("virtualenv --no-site-packages venv", shell=True)
        logger.debug(out)
        out = check_output("venv/bin/pip install -r requirements.txt",
                           shell=True)
        logger.debug(out)
        return True

    def _init_git(self):
        logger.debug("init git")
        try:
            out = check_output("git init && git add . && git commit -m 'init {0}'".
                               format(self.project_name), shell=True)
            logger.debug(out)
        except Exception as e:
            logger.debug(str(e))
            logger.info("git not found, git init ignore")

    @staticmethod
    def _init_dev():
        os.rename("dev.py", "dev")
        os.chmod("dev", 0o755)
        return True

    @staticmethod
    def _init_db():
        shutil.move("app.db", "data/app.db")

    def _init_projects(self):
        time_start = time.time()
        logger.info("init project {0}".format(self.project_name))
        logger.debug(
            "copy from {0} to {1}".format(
                self.template_dir,
                self.work_dir))
        shutil.copytree(self.template_dir, self.work_dir)
        os.chdir(self.work_dir)
        self._make_dirs()
        self._init_config()
        self._init_template()
        self._init_dev()
        self._init_db()
        self._init_git()
        self._init_venv()
        logger.info(
            "init project success using %.3f seconds !!!" %
            (time.time() - time_start))

    def __call__(self, *args, **kwargs):
        self._init_projects()


def do_init(project_name, version, owner, email):
    handler = _InitHandler(project_name, version, owner, email)
    return handler()


# for test
if __name__ == '__main__':
    do_init('data/test{0}'.format(int(time.time())), 'v0.1.1', 'pine', '')



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/5 by allen
"""

import time
from logging import Handler
from redis import StrictRedis, ConnectionPool


class RedisHandler(Handler):
    DEFAULT_HOST = "127.0.0.1"
    DEFAULT_PORT = 6379
    DEFAULT_DB = 0
    DEFAULT_KEY = "logger"

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT,
                 db=DEFAULT_DB, key=DEFAULT_KEY,
                 date_cut=False, conn_pool=None):

        super(RedisHandler, self).__init__()

        self.host = host
        self.port = port
        self.db = db
        self.key = key
        self.date_cut = date_cut
        if conn_pool:
            self.conn_pool = conn_pool
        else:
            self.conn_pool = ConnectionPool(
                host=self.host, port=self.port, db=self.db)

        self.redis = StrictRedis(connection_pool=self.conn_pool)

    def _get_key(self):
        if not self.date_cut:
            return self.key
        else:
            return self.key + ":" + time.strftime("%Y-%m-%d")

    def emit(self, record):
        key = self._get_key()
        msg = self.format(record)
        self.redis.rpush(key, msg)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/19 by allen
"""


from relation import user_role, role_resource
from role import Role
from user import User
from resource import Resource



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from app.lib.database import db, Column
from app.lib.database import MysqlModel as Model
from relation import user_role, role_resource


class Role(Model):

    name = Column(
        db.String(32),
        nullable=False,
        default='-',
        index=True,
        unique=True)
    description = Column(db.String(128), nullable=False, default='-')

    users = db.relationship('User', secondary=user_role,
                            backref=db.backref('roles', lazy='dynamic'),
                            lazy='dynamic')

    resources = db.relationship('Resource', secondary=role_resource,
                                backref=db.backref('roles', lazy='dynamic'),
                                lazy='dynamic')



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/6 by allen
"""

from abc import abstractmethod
import sqlalchemy as sa


class ConversionBase(object):

    @abstractmethod
    def _conversion(self, value):
        raise NotImplementedError

    def conversion_listener(self, target, value, oldvalue, initiator):
        return self._conversion(value)


def conversion_listener(mapper, class_):

    for prop in mapper.iterate_properties:
        try:
            listener = prop.columns[0].type.conversion_listener
        except AttributeError:
            continue
        sa.event.listen(
            getattr(class_, prop.key),
            'set',
            listener,
            retval=True
        )


def auto_conversion(mapper=None):
    if mapper is None:
        mapper = sa.orm.mapper
    sa.event.listen(mapper, 'mapper_configured', conversion_listener)


def inspect_type(mixed):
    if isinstance(mixed, sa.orm.attributes.InstrumentedAttribute):
        return mixed.property.columns[0].type
    elif isinstance(mixed, sa.orm.ColumnProperty):
        return mixed.columns[0].type
    elif isinstance(mixed, sa.Column):
        return mixed.type


def is_case_insensitive(mixed):
    try:
        return isinstance(
            inspect_type(mixed).comparator,
            CaseInsensitiveComparator
        )
    except AttributeError:
        try:
            return issubclass(
                inspect_type(mixed).comparator_factory,
                CaseInsensitiveComparator
            )
        except AttributeError:
            return False


class CaseInsensitiveComparator(sa.Unicode.Comparator):

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveComparator, self).__init__(*args, **kwargs)

    @classmethod
    def lowercase_arg(cls, func):
        def operation(self, other, **kwargs):
            operator = getattr(sa.Unicode.Comparator, func)
            if other is None:
                return operator(self, other, **kwargs)
            if not is_case_insensitive(other):
                other = sa.func.lower(other)
            return operator(self, other, **kwargs)
        return operation

    def in_(self, other):
        if isinstance(other, list) or isinstance(other, tuple):
            other = map(sa.func.lower, other)
        return sa.Unicode.Comparator.in_(self, other)

    def notin_(self, other):
        if isinstance(other, list) or isinstance(other, tuple):
            other = map(sa.func.lower, other)
        return sa.Unicode.Comparator.notin_(self, other)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/6 by allen
"""

import re
import sqlalchemy as sa
from sqlalchemy import types
from app.lib.database.utils import CaseInsensitiveComparator


class EmailType(types.TypeDecorator):

    impl = sa.Unicode
    comparator_factory = CaseInsensitiveComparator
    re_exp = re.compile(r"^\w+([-+.']\w+)*@\w+([-.]\w+)*\.\w+([-.]\w+)*$")

    def __init__(self, length=255, *args, **kwargs):
        super(EmailType, self).__init__(length=length, *args, **kwargs)

    def _validate(self, value):
        if not self.re_exp.match(value):
            raise ValueError('invalid email format %s' % value)
        return value

    def process_bind_param(self, value, dialect):
        if value is not None:
            return self._validate(value.lower())
        return value

    @property
    def python_type(self):
        return self.impl.type.python_type



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/12 by allen
"""


import urllib
from copy import copy
from collections import namedtuple
from flask import request
from sqlalchemy.inspection import inspect
from app.lib.utils import rest_abort
from app.lib.exceptions import ValidationError
from app.lib.database import db, get_model_by_tablename, query_result2dict


_Filter = namedtuple('_Filter', ['key', 'op', 'val'])
_Order = namedtuple('_Order', ['key', 'order'])


class RestQuery(object):

    OPERATORS = ('lt', 'le', 'gt', 'ge', 'eq', 'like', 'in', 'between')

    # use __ prefix distinct with model columns

    PROCESSES = ('__show', '__order')
    PAGINATE = ('__page', '__page_size')

    def __init__(self, app, name, model):
        self.app = app
        self.name = name
        self.model = model
        self.columns = self.model.get_column_names()
        self.relations = inspect(self.model).relationships
        self.default_page_size = self.app.config.get('DEFAULT_PAGE_SIZE', 200)
        self.min_page_size = self.app.config.get('MIN_PAGE_SIZE', 10)
        self.max_page_size = self.app.config.get('MAX_PAGE_SIZE', 1000)
        self.value_separator = self.app.config.get('ITEM_SEPARATOR', ',')
        self.operator_separator = self.app.config.get(
            'OPERATOR_SEPARATOR', '.')
        self.default_order = self.app.config.get('DEFAULT_ORDER', 'desc')
        self.suffix_like = self.app.config.get('SUFFIX_LIKE', True)

    def parse_paginate(self, key, value):

        assert key in self.PAGINATE

        page_error = u'param __page must be a positive integer,' \
                     u'your value is {0}'.format(value)

        page_size_error = u'param __page_size must be a positive ' \
                          u'integer between ({0}, {1}), your value is {2}'.\
            format(self.min_page_size, self.max_page_size, value)

        if key == '__page':
            try:
                page = int(value)
            except ValueError:
                raise ValidationError(page_error)

            if page < 1:
                raise ValidationError(page_error)

            return page
        elif key == '__page_size':
            try:
                page_size = int(value)
            except ValueError:
                raise ValidationError(page_size_error)

            if self.min_page_size < page_size < self.max_page_size:
                return page_size

            raise ValidationError(page_size_error)

    def parse_process(self, key, value):
        assert key in self.PROCESSES

        values = [item.strip() for item in value.split(self.value_separator)]
        if not values:
            raise ValidationError(u'{0} must not empty, your value is {1}'.
                                  format(key, value))

        if key == '__show':
            invalid_item = set(self.columns) - set(values)
            if invalid_item:
                raise ValidationError(u'{0} of your __show columns {1} is not'
                                      u' an attribute of {2}  '.format(
                                          u','.join(invalid_item), value, self.name))

            return values
        elif key == '__order':
            order_by_columns = []
            for val in values:
                if val.endsith(('.desc', '.asc')):
                    attr, order = val.split('.')
                else:
                    attr, order = val, self.default_order

                if attr not in self.columns:
                    raise ValidationError(u'{0}  of your __order_by columns '
                                          u'is not an attribute of {1}'.format(
                                              attr, value, self.name))

                order_by_columns.append(_Order(key=attr, order=order))

            return order_by_columns

    def parse_filter(self, key, value):
        if key in self.columns:
            return _Filter(key=key, op='eq', val=value)
        elif key.endswith(self.OPERATORS):
            key_oper = key.split(self.operator_separator)
            if len(key_oper) != 2:
                raise ValidationError(u'invalid filter  {0} for {1}'.format(
                    key, self.name))

            fkey, oper = key_oper

            if fkey not in self.columns:
                raise ValidationError(u'filter key {0} is not an attribute of {1}'.format(
                    fkey, self.name
                ))
            elif oper not in self.OPERATORS:
                raise ValidationError(
                    u'filter operation {0} is not support'.format(oper))

            if oper == 'like' and not self.suffix_like and value.startswith(
                    '%'):
                raise ValidationError(u'suffix like is not support, you used at {0}={1}'.format(
                    key, value))
            elif oper in ['in', 'between']:
                value = value.split(self.value_separator)
                if oper == 'between' and len(value) != 2:
                    raise ValidationError(u'between value must be a range,'
                                          u' your query: {0}={1}'.format(
                                              key, self.value_separator.join(value)))

            return _Filter(key=fkey, op=oper, val=value)

        raise ValidationError(u'invalid filter {0} for {1}'.format(
            key, self.name))

    @staticmethod
    def get_show_query(model, show_columns):
        if show_columns:
            show_columns = [getattr(model, column) for column in show_columns]
            return db.session.query(show_columns)
        else:
            return db.session.query(model)

    @staticmethod
    def order_query(model, query, orders):
        for order in orders:
            if order.order == 'desc':
                query = query.order_by(getattr(model, order.key))
            else:
                query = query.order_by(-getattr(model, order.key))

        return query

    @staticmethod
    def paginate_query(query, page, page_size):
        offset = (page - 1) * page_size
        return query.offset(offset).limit(page_size)

    @staticmethod
    def filter_query(model, query, filters):

        for query_filter in filters:
            filter_column = getattr(model, query_filter.key)
            if query_filter.op == 'eq':
                query = query.filter(filter_column == query_filter.val)
            elif query_filter.op == 'lt':
                query = query.filter(filter_column < query_filter.val)
            elif query_filter.op == 'le':
                query = query.filter(filter_column <= query_filter.val)
            elif query_filter.op == 'gt':
                query = query.filter(filter_column > query_filter.val)
            elif query_filter.op == 'ge':
                query = query.filter(filter_column >= query_filter.val)
            elif query_filter.op == 'like':
                query = query.filter(filter_column.like(query_filter.val))
            elif query_filter.op == 'in':
                query = query.filter(filter_column.in_(query_filter.val))
            elif query_filter.op == 'between':
                query = query.filter(filter_column.between(*query_filter.val))

        return query

    def generate_query(self, filters, processes, paginate):
        show_query = self.get_show_query(
            self.model, processes.get('__show', []))
        filter_query = self.filter_query(self.model, show_query, filters)
        paginate_query = self.paginate_query(
            filter_query, paginate['__page'], paginate['__page_size'])
        return paginate_query

    def generate_relation_query(
            self, filters, processes, paginate, rel_name, pk):

        relation = self.relations[rel_name]
        relation_model = get_model_by_tablename(relation.table.fullname)
        show_query = self.get_show_query(
            relation_model, processes.get('__show', []))
        rel_query = show_query.filter(
            getattr(
                relation_model,
                relation.back_populates).any(
                id=pk))
        filter_query = self.filter_query(relation_model, rel_query, filters)
        paginate_query = self.paginate_query(
            filter_query, paginate['__page'], paginate['__page_size'])

        return paginate_query

    @staticmethod
    def get_prev_link(params, paginate):
        prev_params = copy(params)
        prev_params['__page'] = paginate['__page'] - 1
        if prev_params['__page'] >= 1:
            return request.base_url + '?' + urllib.urlencode(prev_params)
        else:
            return None

    @staticmethod
    def get_next_link(params, paginate):
        prev_params = copy(params)
        prev_params['__page'] = paginate['__page'] + 1
        return request.base_url + '?' + urllib.urlencode(prev_params)

    def parse_params(self, params):
        processes = dict()
        paginate = dict()
        filters = []

        params.setdefault('__page', 1)
        params.setdefault('__page_size', self.default_page_size)
        for key, value in params.items():
            if key in self.PROCESSES:
                processes[key] = self.parse_process(key, value)
            elif key in self.PAGINATE:
                paginate[key] = self.parse_paginate(key, value)
            else:
                filters.append(self.parse_filter(key, value))

        return processes, filters, paginate

    def query(self, params):
        rel_name = params.pop('__relation', None)
        pk = params.pop('__pk', None)

        processes, filters, paginate = self.parse_params(params)
        if rel_name and pk:
            if rel_name not in self.relations:
                raise ValidationError(u'{0} is not a relation of {1}'.format(
                    rel_name, self.name))
            query = self.generate_relation_query(
                filters, processes, paginate, rel_name, pk)
        else:
            query = self.generate_query(filters, processes, paginate)
        result = [query_result2dict(item) for item in query]

        return {
            'result': result,
            'page': paginate['__page'],
            'page_size': paginate['__page_size'],
            'prev_page': self.get_prev_link(params, paginate),
            'next_page': self.get_next_link(params, paginate)
        }

    def __call__(self, params):
        return self.query(params)


class RestModel(object):

    def __init__(self, app, name, model=None):
        self.app = app
        self.name = name
        self.model = model or get_model_by_tablename(name)
        self.rest_query = RestQuery(
            app=self.app, name=self.name, model=self.model)
        self.columns = self.rest_query.columns
        self._url_map = dict()

    def validate_keys(self, keys):
        invalid_keys = frozenset(keys) - frozenset(self.columns)

        if invalid_keys:
            raise ValidationError(u'{0} is not attributes of {1}'.format(
                u','.join(invalid_keys), self.name))

        return True

    def http_get(self, params):
        return self.rest_query(params)

    def http_get_pk(self, params):
        item = self.model.get(params['id'])
        return item.to_dict() if item else rest_abort(404)

    def http_post(self, data):
        if data.pop('id', None):
            raise ValidationError(u"id can't be specified when post")

        self.validate_keys(data.keys())
        item = self.model(**data)
        return item.save().to_dict()

    def http_delete(self, params):
        # if the item need to delete not exists,
        # return empty dict, else return the item dict

        item = self.model.get(params['id'])
        res = dict()
        if item:
            res = item.to_dict()
            item.delete()
        return res

    def http_put(self, data):
        _id = data.pop('id')
        self.validate_keys(data.keys())

        if not self.model.get(_id):
            raise ValidationError(
                u'id {0} of {1} not exists'.format(
                    _id, self.name))

        self.model.query_update(query={'id': _id}, data=data)
        return self.model.get(_id).to_dict()

    def init_bp(self, bp):
        bp.rest_route(rule=self.name, methods=['GET'])(
            self.http_get, '_'.join((self.name, self.http_get.__name__)))

        bp.rest_route(rule=self.name + '/<id>', methods=['GET'])(
            self.http_get_pk, '_'.join((self.name, self.http_get_pk.__name__)))

        bp.rest_route(rule=self.name + '/<id>', methods=['PUT'])(
            self.http_put, '_'.join((self.name, self.http_put.__name__)))

        bp.rest_route(rule=self.name, methods=['POST'])(
            self.http_post, '_'.join((self.name, self.http_post.__name__)))

        bp.rest_route(rule=self.name + '/<id>', methods=['DELETE'])(
            self.http_delete, '_'.join((self.name, self.http_delete.__name__)))

        bp.rest_route(rule=self.name + '/<__pk>/<__relation>', methods=['GET'])(
            self.http_get, '_'.join(
                (self.name, self.http_get.__name__, 'relations'))
        )

        return bp



#!/usr/bin/env python
# chkconfig: - 30 74
# description:  This is a daemon which handles app

"""
    create at 2017/11/24 by allen
"""

import os
import sys
import imp
import time
import signal
from subprocess import check_output, CalledProcessError

SERVICE_NAME = "{{template}}"
WORK_DIR = os.path.curdir
DEBUG = False
GUN_CONF = "gun.py"
APP = "main:app"
STOP_WAIT = 5
START_WAIT = 20
Red = '\033[0;31m'
Green = '\033[0;32m'
Yellow = '\033[0;33m'
Color_Off = '\033[0m'


def log_error(msg):
    print(Red + msg + Color_Off)


def log_info(msg):
    print(Green + msg + Color_Off)


def log_debug(msg):
    if DEBUG:
        print(Yellow + msg + Color_Off)


def log_fatal(msg):
    log_error(msg)
    exit(-1)


def get_gun_module():
    try:
        _gun = imp.load_source("gun", GUN_CONF)
        return _gun
    except Exception as e:
        log_fatal(
            "file {0} must exists, current error {1}".format(
                GUN_CONF, str(e)))


gun = get_gun_module()


def get_pidfile_path():
    try:
        return gun.pidfile
    except Exception as e:
        log_fatal("file {0} must exists and has pidfile conf item, current error: {1}".
                  format(GUN_CONF, str(e)))


pid_file = get_pidfile_path()


def execute_cmd(cmd):
    try:
        out = check_output(cmd, shell=True)
        return True, out
    except CalledProcessError as e:
        log_debug("execute {0} failed, {1}".format(cmd, str(e)))
        return False, e.message


def read_pid():
    try:
        with open(pid_file, 'r') as pfp:
            return int(pfp.read())
    except Exception as e:
        log_debug(str(e))
        return -1


def get_pid_sub(pid):
    ps_cmd = "ps -eo ppid,pid,start | grep {0}".format(pid)
    ps_status, ps_output = execute_cmd(ps_cmd)
    if not ps_status:
        return None, None

    master = ()
    workers = []
    for line in ps_output.splitlines():
        cur_ppid, cur_pid, cur_start = line.split()
        proc = (int(cur_pid), cur_start)
        if proc[0] == pid:
            master = proc
        else:
            workers.append(proc)

    return master, workers


def do_status():
    pid = read_pid()
    if pid < 0:
        log_info("service {0} stopped".format(SERVICE_NAME))
        return False

    master, workers = get_pid_sub(pid)
    if not master or not workers:
        log_info("service {0} stopped".format(SERVICE_NAME))
        return False

    log_info("master process pid: {master_id}, started at {master_started};".format(
        master_id=pid, master_started=master[1]))
    log_info("worker process count: %d" % len(workers))

    for worker in sorted(workers):
        log_info("pid {0}, started at {1};".format(worker[0], worker[1]))

    return True


def service_alive():
    pid = read_pid()
    if pid < 0:
        return False

    master, workers = get_pid_sub(pid)
    if master or workers:
        return True

    return False


def clear_pidfile():
    try:
        os.remove(pid_file)
    except Exception as e:
        log_debug(str(e))
        pass


def do_start():

    start_time = time.time()
    if service_alive():
        log_fatal(
            "service {0} has been started, can't be start twice".format(SERVICE_NAME))

    clear_pidfile()
    gun_cmd = "venv/bin/python venv/bin/gunicorn -c {0} {1}".format(
        GUN_CONF, APP)
    status, output = execute_cmd(gun_cmd)
    if not status:
        log_fatal(
            "execute {0} failed, start {1} failed".format(
                gun_cmd, SERVICE_NAME))

    if not wait_start():
        log_fatal("wait start  %s  for %.3f seconds failed" %
                  (SERVICE_NAME, time.time() - start_time))

    if not do_status():
        log_fatal("start {0} failed".format(SERVICE_NAME))

    log_info("start %s success using %.3f seconds" %
             (SERVICE_NAME, time.time() - start_time))
    return True


def kill_process(pid, sig):
    try:
        os.kill(pid, sig)
    except Exception as e:
        log_debug('kill {0} failed, {1}'.format(pid, str(e)))
        return False

    return True


def wait_start():
    return wait_status(START_WAIT, True)


def wait_stop():
    return wait_status(STOP_WAIT, False)


def wait_status(wait_seconds, alive):
    wait_secs = 0
    while wait_secs <= wait_seconds:

        if service_alive() == alive:
            return True

        time.sleep(.1)

    return service_alive() == alive


def do_stop():
    start_time = time.time()

    if not service_alive():
        clear_pidfile()
        log_info("service {0} has been stopped".format(SERVICE_NAME))
        return True

    pid = read_pid()
    if not kill_process(pid, signal.SIGTERM):  # graceful stop
        log_fatal(
            "kill process {0} failed, stop {1} failed".format(
                pid, SERVICE_NAME))

    if not wait_stop():
        kill_process(read_pid(), signal.SIGQUIT)  # quick stop

    if not wait_stop():
        log_fatal("pid file {0}  still exists, stop {1} failed".format(
            pid_file, SERVICE_NAME))

    # pid file removed
    master, workers = get_pid_sub(pid)
    if master or workers:
        processes = workers.append(master)
        log_fatal('process {0} still exists, stop {1}  failed'.format(
            ','.join([proc[0] for proc in processes]), SERVICE_NAME))

    log_info("stop %s success using %.3f seconds" %
             (SERVICE_NAME, time.time() - start_time))
    return True


def do_reload():
    start_time = time.time()

    if not service_alive():
        log_info(
            "service {0} has been stopped, start service.".format(SERVICE_NAME))
        return do_start()

    pid = read_pid()
    if not kill_process(pid, signal.SIGHUP):
        log_fatal("reload {0} failed".format(SERVICE_NAME))

    time.sleep(.5)

    if not do_status():
        log_fatal("reload {0} failed".format(SERVICE_NAME))

    log_info("reload %s success using %.3f seconds" %
             (SERVICE_NAME, time.time() - start_time))


def do_help():
    help_map = {
        'status': 'show status of service {0}'.format(SERVICE_NAME),
        'start': 'start service {0}'.format(SERVICE_NAME),
        'stop': 'stop service {0}'.format(SERVICE_NAME),
        'reload': 'reload config of service {0}, if not started, do start'.format(SERVICE_NAME),
        'help': 'show usage of this script'
    }

    for key, value in help_map.items():
        log_info("{0}: {1}".format(key, value))

    exit(0)


if __name__ == '__main__':
    action_map = {
        'status': do_status,
        'start': do_start,
        'stop': do_stop,
        'reload': do_reload,
        'help': do_help
    }

    if len(sys.argv) == 1:
        action = 'help'
    else:
        action = sys.argv[1]

    if action not in action_map:
        do_help()

    action_map[action]()
    exit(0)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/5 by allen
"""

import socket
import logging
import logging.handlers as sys_handlers

from flask import request

import handlers as custom_handlers


class CustomFormatter(logging.Formatter):

    def format(self, record):
        try:
            setattr(record, 'url', request.url)
        except RuntimeError:
            setattr(record, 'url', 'Null')

        setattr(record, 'hostname', socket.gethostname())
        return logging.Formatter.format(self, record)


class AppLogger(object):

    def __init__(self, app):
        self._app = app
        self._default_formatter = self._app.config.get('LOG_FORMATTER')
        self._default_level = self._app.config.get('LOG_LEVEL', logging.INFO)

    def init_handler(self, handler_cls, config):
        formatter = config.pop('formatter', self._default_formatter)
        level = config.pop('level', self._default_level)
        handler = handler_cls(**config)
        if formatter:
            handler.setFormatter(CustomFormatter(formatter))
        handler.setLevel(level)
        self._app.logger.addHandler(handler)
        return handler

    @staticmethod
    def _get_handler_cls(handler_name):
        if not handler_name:
            raise Exception(u'logger handler is required')

        cls = getattr(logging, handler_name, None) or getattr(sys_handlers, handler_name, None) \
            or getattr(custom_handlers, handler_name, None)

        if cls and issubclass(cls, logging.Handler):
            return cls

        raise Exception(u'handler {0} is not support'.format(handler_name))

    def __call__(self, *args, **kwargs):
        for handler in self._app.logger.handlers:
            handler.setFormatter(CustomFormatter(self._default_formatter))

        for logger_name, logger_config in self._app.config['LOGGERS'].items():
            self._app.logger.debug(u'init logger {0}'.format(logger_name))

            if logger_config.pop('disable', None):
                continue

            handler_name = logger_config.pop('handler', None)
            handler_cls = self._get_handler_cls(handler_name)

            self.init_handler(handler_cls, logger_config)

        return self._app.logger


def init_logger(app):
    logger = AppLogger(app)()
    return logger



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from collections import OrderedDict
from flask import current_app
from app.lib.exceptions import ValidationError, InternalError


class Validator(object):

    def __init__(self, validator):
        self.validator = validator
        self.validate_chain = OrderedDict([
            ('required', self.process_required),
            ('nonempty', self.process_nonempty),
            ('types', self.process_types),
            ('oneof', self.process_oneof),
            # can only do unique for list of hashable objects
            ('unique', self.process_unique),
            ('length', self.process_length),
            ('override', self.process_override),
            ('default', self.process_default),
            ('custom', self.process_custom)  # validate data logic
        ])

    def process_required(self, data):
        required = self.validator['required']
        not_supply = frozenset(required) - frozenset(data.keys())
        if not_supply:
            raise ValidationError(
                u'param %s is required' %
                u','.join(not_supply))

        return True

    def process_nonempty(self, data):
        no_empty = self.validator['nonempty']
        for k in no_empty:
            if not data.get(k, None):
                raise ValidationError(
                    u'param {0} required not empty'.format(k))

        return True

    def process_types(self, data):
        types = self.validator['types']

        for key, tp in types.items():
            if key in data and not isinstance(data[key], tp):
                raise ValidationError(u'param {0} need to be type {1}, your type is {2}'
                                      .format(key, tp, type(data[key])))

        return True

    def process_unique(self, data):
        unique_params = self.validator['unique']

        for k in unique_params:
            if k in data:
                data[k] = list(frozenset(k))

        return True

    def process_length(self, data):
        length = self.validator['length']
        for key, ll in length.items():
            if key not in data:
                continue

            val_len = len(data[key])

            if val_len < ll[0] or val_len > ll[-1]:
                raise ValidationError(u'param {0} length required to between {1} and {2}, '
                                      u'your length is {3}'.format(key, ll[0], ll[-1], val_len))

        return True

    def process_oneof(self, data):
        oneof = self.validator['oneof']

        for key, option_values in oneof.items():
            if key in data and data[key] not in option_values:
                raise ValidationError(u'param {0} must in {1}, your value is {2}'.format(
                    key, u','.join(option_values), data[key]
                ))

        return True

    def process_override(self, data):
        override = self.validator['override']
        for key, override_fn in override.items():
            if key in data:
                try:
                    data[key] = override_fn(key, data[key])
                except Exception as e:
                    current_app.logger.exception(u'process override with key={0}, value={1}, error: {2}'.format(
                        key, data[key], str(e)))

                    raise InternalError(
                        u'oops, validate param occurs internal error')

        return True

    def process_default(self, data):
        default_args = self.validator['default']

        for k, v in default_args.items():
            data.setdefault(k, v)

        return True

    def process_custom(self, data):
        custom = self.validator['custom']

        for validate_fn in custom:
            validate_fn(data)

    def __call__(self, data):
        for validate_name, validate_fn in self.validate_chain.items():
            if validate_name in self.validator:
                validate_fn(data)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""


class CustomError(Exception):

    def __init__(self, msg):
        super(CustomError, self).__init__(msg)
        self.msg = msg
        self.code = 500

    def to_dict(self):
        return dict(code=self.code, msg=self.msg)

    def __unicode__(self):
        return unicode(self.msg)

    def __str__(self):
        return str(self.msg)


class InternalError(CustomError):

    def __init__(self, msg):
        super(InternalError, self).__init__(msg)
        self.code = 500


class ValidationError(CustomError):

    def __init__(self, msg):
        super(ValidationError, self).__init__(msg)
        self.code = 400


class SupportError(InternalError):
    def __init__(self, msg):
        super(SupportError, self).__init__(msg)
        self.code = 500


class AuthError(CustomError):

    def __init__(self, msg):
        super(AuthError, self).__init__(msg)
        self.code = 401


class PermissionError(CustomError):
    def __init__(self, msg):
        super(PermissionError, self).__init__(msg)
        self.code = 403


class RestHttpError(CustomError):

    _code_map = {
        400: 'bad request, please check your params ',
        401: 'authenticate failed, please check your token or password',
        403: 'permission denied, you has not permission access this resource',
        404: 'resource not found, please check your url',
        500: 'sorry, internal error happened',
        405: 'request method not allowed'
    }

    def __init__(self, code):
        msg = self._code_map[code]
        super(RestHttpError, self).__init__(msg)
        self.code = code



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/2 by allen
"""

import logging
from logging import StreamHandler, INFO
from termcolor import colored


_logger = logging.getLogger(__name__)
_logger.setLevel(INFO)
_logger.addHandler(StreamHandler())


def waring(msg):
    return _logger.warning(colored(msg, "yellow"))


def error(msg):
    return _logger.error(colored(msg, "red"))


def debug(msg):
    return _logger.debug(colored(msg, "green"))


def info(msg):
    return _logger.info(colored(msg, "green"))


def set_level(level):
    return _logger.setLevel(level)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/5 by allen
"""

from flask import request, session
from sqlalchemy.exc import SQLAlchemyError

from app.lib.logger import init_logger
from app.lib.utils import flask_res
from app.lib.exceptions import CustomError
from app.lib.pine_wrapper import PineBlueprint
from app.lib.rest_model import RestModel
from app.lib.utils import get_bp_prefix
from app.lib.auth import PermissionAuth


class Init(object):
    def __init__(self, app, bp_list=None):
        self._app = app
        self._bp_list = bp_list or []
        self.permission_auth_handler = PermissionAuth()

    def __call__(self, *args, **kwargs):
        self.init_config()
        self.init_logger()
        self.init_before()
        self.init_views()
        self.init_extensions()
        self.init_error_handler()
        self.init_after()
        self.init_rest_db()
        return self._app

    def init_logger(self):
        self._app.logger.debug('init logger')
        init_logger(self._app)

    def init_config(self):
        self._app.logger.debug('init config')
        self._app.config.from_object('config')
        self._app.config.from_pyfile('config.py')

    def init_before(self):
        self._app.logger.debug('init before_request')

        @self._app.before_request
        def before_request():
            ip_list = request.headers.getlist("X-Forwarded-For")
            session['remote_address'] = ip_list[0].split(
                ',')[0] if ip_list else request.remote_addr
            self.permission_auth_handler.validate_request_permission()

    def init_after(self):
        self._app.logger.debug('init after_request')

        @self._app.after_request
        def after_request(response):
            response.headers['Server'] = self._app.config['APP_NAME']
            return response

    def init_error_handler(self):
        self._app.logger.debug("init error handler")

        @self._app.errorhandler(CustomError)
        def handler_custom_error(error):
            return flask_res(data=error, code=error.code)

        @self._app.errorhandler(SQLAlchemyError)
        def handler_db_error(error):
            msg = u'DB操作失败: {0}'.format(str(error))
            self._app.logger.exception(msg)
            data = {'code': 500, 'msg': msg}
            return flask_res(data=data, code=500)

    def init_extensions(self):
        from app import extensions
        self._app.logger.debug("init extensions")
        module_priority_chain = getattr(
            extensions, 'module_priority_chain', [])

        for name in module_priority_chain:
            try:
                _module = __import__(
                    '{0}.{1}'.format(
                        extensions.__name__,
                        name),
                    fromlist=[''])
                if hasattr(_module, 'init_app'):
                    init_fn = getattr(_module, 'init_app', None)
                elif hasattr(_module, 'extension'):
                    init_fn = getattr(_module.extension, 'init_app', None)
                if not init_fn:
                    raise Exception(
                        u'init_app of {0} is not defined'.format(name))
                elif not callable(init_fn):
                    raise Exception(
                        u'init_app of {0} is not callable'.format(name))

                init_fn(self._app)
            except ImportError as e:
                self._app.logger.exception(str(e))

    def register_bp(self, bp):
        bp_prefix = get_bp_prefix(self._app, bp.name)
        self._app.register_blueprint(bp, url_prefix=bp_prefix)

    def init_views(self):
        import views

        self._app.logger.debug("init views")

        for bp in self._bp_list:
            self.register_bp(bp)

    def init_rest_db(self):
        from app import models
        from app.lib.database import get_model_registry

        if not self._app.config.get('REST_DB'):
            self._app.logger.debug("ignore init rest_db")
            return True

        self._app.logger.debug("init rest_db")
        rest_db = PineBlueprint('rest_db', __name__)

        rest_models = (RestModel(self._app, model.__tablename__, model)
                       for model in get_model_registry())

        for rest_model in rest_models:
            rest_model.init_bp(bp=rest_db)

        self.register_bp(rest_db)

        return True


def init_app(app, bp_list):
    with app.app_context():
        return Init(app, bp_list)()



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/6 by allen
"""

from ipaddress import ip_address
from sqlalchemy import types
from app.lib.database.utils import ConversionBase


class IpAddressType(types.TypeDecorator, ConversionBase):
    impl = types.BigInteger

    def __init__(self, *args, **kwargs):

        super(IpAddressType, self).__init__(*args, **kwargs)
        self.impl = types.BigInteger

    def process_bind_param(self, value, dialect):
        if isinstance(value, basestring):
            return int(ip_address(value))
        else:
            return int(value)

    def process_result_value(self, value, dialect):
        return ip_address(value) if value else None

    def _conversion(self, value):
        return ip_address(value) if value else None

    @property
    def python_type(self):
        return self.impl.type.python_type



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/4 by allen
"""

import os
from binascii import hexlify
from base64 import b64encode

VERSION = "1.0.8"

TEMPLATE_SRC = 'template/'
TEMPLATE_HEADER = "#!/usr/bin/env python\n# coding: utf-8\n"
TEMPLATE_DIRS_CREATE = ('data/', 'logs/', 'instance/')
TEMPLATE_CONF_NAME = "config.py"
TEMPLATE_GUN_CONF = "gun.py"
TEMPLATE_SECRET_KEY = b64encode(hexlify(os.urandom(32))).decode("utf-8")

# project config, render to config.py
TEMPLATE_LOG_FORMAT = '%(hostname)s %(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s '

TEMPLATE_PROJECT = {
    '__KEYS_ORDER': ['APP_NAME', 'VERSION', 'OWNER', 'EMAIL', 'DEBUG', '\n',
                     'LOG_FORMATTER', 'LOGGERS', 'REST_DB', 'SQLALCHEMY_TRACK_MODIFICATIONS'],

    'VERSION': "v0.0.1",
    'OWNER': "pine",
    'DEBUG': False,
    'EMAIL': '',
    'REST_DB': True,
    'SQLALCHEMY_TRACK_MODIFICATIONS ': False,
    'LOG_FORMATTER': TEMPLATE_LOG_FORMAT,
    'LOGGERS': {
        'error_log': {
            'handler': 'RotatingFileHandler',
            'level': 'ERROR',
            'formatter': TEMPLATE_LOG_FORMAT,
            'filename': 'logs/error.log',
            'maxBytes': 1024 * 1024 * 100,
            'backupCount': 7,
            'encoding': 'utf-8'
        },

        'app_log': {
            'handler': 'RotatingFileHandler',
            'level': 'INFO',
            'formatter': TEMPLATE_LOG_FORMAT,
            'filename': 'logs/app.log',
            'maxBytes': 1024 * 1024 * 100,
            'backupCount': 7,
            'encoding': 'utf-8'
        },

        'mail_log': {
            'handler': 'SMTPHandler',
            'level': 'ERROR',
            'formatter': TEMPLATE_LOG_FORMAT,
            'fromaddr': 'error@{app_name}.com',
            'toaddrs': ['{email}'],
            'mailhost': 'smtp.{app_name}.com',
            'disable': 1,  # # use number for boolean, compatible with json
        },

        "redis_log": {
            "handler": "RedisHandler",
            "host": "127.0.0.1",
            "port": 6379,
            "date_cut": 1,
            "db": 0,
            "key": "{app_name}:log",
            "level": "INFO",
            "disable": 1
        }

    }
}

TEMPLATE_INSTANCE = {
    '__KEYS_ORDER': ['DEBUG', 'SECRET_KEY', '\n',
                     'SQLALCHEMY_DATABASE_URI', 'SQLALCHEMY_ECHO',
                     'SQLALCHEMY_RECORD_QUERIES',
                     'REDIS_CONFIG'],

    'DEBUG': True,
    'SECRET_KEY': TEMPLATE_SECRET_KEY,
    'SQLALCHEMY_DATABASE_URI': 'sqlite:///data/app.db',
    'SQLALCHEMY_ECHO': True,
    'SQLALCHEMY_RECORD_QUERIES': True,

    'REDIS_CONFIG': {
        "host": "localhost",
        "port": 6379,
        "db": 1
    }
}

# gunicorn config, render to gun.py
TEMPLATE_GUNICORN = {
    "__KEYS_ORDER": ['bind', 'workers', 'worker_class', 'backlog', 'daemon', 'pidfile',
                     'loglevel', 'accesslog', 'errorlog', 'access_log_format'],

    "bind": "0.0.0.0:8080",
    "workers": "2*cpu",
    "worker_class": "sync",
    "backlog": 2048,
    "daemon": True,
    "loglevel": "info",
    "pidfile": "data/gun.pid",
    "accesslog": 'logs/access.log',
    "errorlog": 'logs/error.log',
    "access_log_format": "%({X-Real-IP}i)s %(h)s %(l)s %(u)s %(t)s %(r)s %(s)s %(b)s %(f)s %(a)s"
}



#!venv/bin/python
# coding: utf-8

"""
    create at 2017/11/4 by allen
"""

from flask_script import Manager, Shell
from flask_script.commands import ShowUrls, Clean

from app import models
from app.lib.database import db
from main import app

manager = Manager(app)


def _make_context():
    return dict(app=app, db=db)


@manager.command
def create_db():
    models.create_db()


manager.add_command('url', ShowUrls())
manager.add_command('clean', Clean())
manager.add_command("shell", Shell(make_context=_make_context))


if __name__ == '__main__':
    manager.run()



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/2 by allen
"""

from logging import DEBUG
import click

import tadpole.logger as logger
from tadpole.config import TEMPLATE_PROJECT, VERSION
from tadpole.init_handler import do_init


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(VERSION)
    ctx.exit()


def open_debug(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    logger.set_level(DEBUG)
    logger.debug('debug mod opened')


@click.group()
@click.option('-v', '--version', is_flag=True, callback=print_version,
              expose_value=False, is_eager=True)
def cli():
    pass


@cli.command()
@click.option('-d', '--debug', is_flag=True, callback=open_debug,
              expose_value=False, is_eager=True)
@click.option('-n', '--name', prompt=True, help="project name")
@click.option('-v', '--version',
              default=TEMPLATE_PROJECT['VERSION'], help="project version")
@click.option('-o', '--owner',
              default=TEMPLATE_PROJECT['OWNER'], help="project owner")
@click.option('-e', '--email',
              default=TEMPLATE_PROJECT['EMAIL'], help="project email")
def init(name, version, owner, email):
    logger.debug("raw input project name: {0}, version {1}, owner: {2}, email: {3}".format(
        name, version, owner, email
    ))

    do_init(name, version, owner, email)


if __name__ == '__main__':
    cli()



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/19 by allen
"""

from sqlalchemy import Index
from app.lib.database import db, Column
from app.lib.database import MysqlModel as Model


class Resource(Model):
    rtype = Column(db.String(32), nullable=False, default='-')
    name = Column(db.String(128), nullable=False, default='-')
    operation = Column(db.String(32), nullable=False, default='-')
    description = Column(db.String(128), nullable=False, default='-')


Index('name_operation_uc', Resource.name, Resource.operation, unique=True)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/22 by allen
"""

import re
from flask import request, session, current_app
from app.lib.constant import ResourceType
from app.models.auth import Resource, role_resource, Role, user_role, User
from app.lib.exceptions import AuthError, PermissionError


class HttpBasicAuth(object):

    def __init__(self, user_loader, hash_password_handler=None,
                 verify_password_handler=None):

        self.user_loader = user_loader
        self.hash_password_handler = hash_password_handler
        self.verify_password_handler = verify_password_handler

    def hash_password(self, auth):
        if self.hash_password_handler:
            try:
                return self.hash_password_handler(auth.password)
            except Exception as e:
                current_app.logger.exception(str(e))

            try:
                return self.hash_password_handler(auth.username, auth.password)
            except Exception as e:
                current_app.logger.exception(str(e))

        return auth.password

    def get_user(self, auth):
        if not auth or not auth.username:
            return None

        user = self.user_loader(auth.username)

        return user

    def auth_user(self, auth):
        if session.get('user_account'):
            return self.user_loader(session['user_account'])

        user = self.get_user(auth)
        stored_password = user.password if user else None
        if not stored_password:
            return None

        if self.verify_password_handler:
            return self.verify_password_handler(auth.username, auth.password)

        client_password = self.hash_password(auth)
        if stored_password == client_password:
            session['user_account'] = user.account
            return user

        return None

    def __call__(self, auth):
        return self.auth_user(auth=auth)


class AuthLoaderBase(object):

    @staticmethod
    def get_user_resources(user_id):
        raise NotImplemented

    @staticmethod
    def load_user(account):
        raise NotImplemented

    @staticmethod
    def get_user_resources(user_id):
        raise NotImplemented

    @staticmethod
    def load_resources(rtype, name, operation):
        raise NotImplemented


class AuthDbLoader(AuthLoaderBase):

    @staticmethod
    def get_user_resources(user_id):
        return Resource.query.join(role_resource, Role, user_role, User). \
            filter(User.id == user_id)

    @staticmethod
    def load_user(account):
        return User.get_by(account=account).first()

    @staticmethod
    def get_user_resources(user_id):
        return Resource.query.join(role_resource, Role, user_role, User). \
            filter(User.id == user_id)

    @staticmethod
    def load_resources(rtype, name, operation):
        http_resources = Resource.get_by(rtype=rtype)

        return (resource for resource in http_resources
                if re.match(resource.name, name) and
                operation in resource.operation.split(','))

    @staticmethod
    def load_user_roles(user_id):
        return Role.query.join(user_role).filter(user_id=user_id)


_auth_db_loader = AuthDbLoader()

_http_basic_auth = HttpBasicAuth(user_loader=_auth_db_loader.load_user)


class PermissionAuth(object):

    def __init__(self, http_auth_handler=_http_basic_auth,
                 auth_info_loader=_auth_db_loader):
        self.auth_info_loader = auth_info_loader

        self.auth_handler = http_auth_handler

    def validate_user_permission(self, user, resources):
        user_resources = set(resource.id for
                             resource in self.auth_info_loader.
                             get_user_resources(user.id))

        access_resources = set(resource.id for resource in resources)

        if access_resources.issubset(user_resources):
            return True

    def is_root_user(self, user):
        roles = self.auth_info_loader.load_user_roles(user.id)
        return any(role.name == 'root' for role in roles)

    def validate_request_permission(self):
        path_resources = list(self.auth_info_loader.load_resources(
            ResourceType.HTTP, request.path, request.method.upper()))
        if not path_resources:
            return True
        user = self.auth_handler(request.authorization)
        if not user:
            raise AuthError(u'authenticate failed,'
                            u' please check your username or password')

        # super admin, ignore permission
        if self.is_root_user(user):
            return True

        # validate permission
        if not self.validate_user_permission(user, path_resources):
            raise PermissionError(u'permission denied, your  have not'
                                  u' permission to do {0} on {1}'.format(
                                      request.path, request.method.upper()))



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/6 by allen
"""

from email import EmailType as Email
from ip_address import IpAddressType as IpAddress
from password import PasswordType as Password

__all__ = (
    'Email', 'IpAddress', 'Password'
)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from datetime import datetime

from app.extensions.sqlalchemy import extension as db
from app.lib.utils import get_relation_url
from sqlalchemy import inspect

import custom_types

Column = db.Column


class CRUDMixin(object):

    @classmethod
    def create(cls, **kwargs):
        instance = cls(**kwargs)
        return instance.save()

    @classmethod
    def query_update(cls, query=None, data=None):
        query = query or dict()
        update_query = getattr(cls, 'query').filter_by(**query)
        update_query.update(data or dict())
        db.session.commit()
        return update_query.all()

    def save(self, commit=True):
        db.session.add(self)
        if commit:
            db.session.commit()
        return self

    def delete(self, commit=True):
        db.session.delete(self)
        return commit and db.session.commit()


class SurrogatePK(object):

    __table_args__ = {'extend_existing': True}

    @classmethod
    def get(cls, record_id):
        data = getattr(cls, 'query').get(record_id)
        return data

    @classmethod
    def get_by(cls, **kwargs):
        return getattr(cls, 'query').filter_by(**kwargs)


class Model(SurrogatePK, db.Model, CRUDMixin):
    __abstract__ = True

    id = Column(db.Integer, nullable=False, primary_key=True)
    create_time = Column(db.DateTime, nullable=False,
                         default=datetime.now, onupdate=datetime.now)

    @classmethod
    def get_columns(cls):
        return cls.__table__.columns

    @classmethod
    def get_column_names(cls):
        return [c.name for c in cls.get_columns()]

    def _as_dict(self):  # for common sqlalchemy  query result to dict
        from main import app
        hide_attrs = getattr(self, '__hide__', ())

        column_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns
                       if c.name not in hide_attrs}

        try:
            for rel_name, _ in inspect(type(self)).relationships.items():
                if rel_name in hide_attrs:
                    continue

                column_dict['__' + rel_name + '_link'] = get_relation_url(
                    self.__tablename__, self.id, rel_name)

        except Exception as e:
            app.logger.info(str(e))

        return column_dict

    def to_dict(self):  # for human
        return self._as_dict()


class MysqlModel(Model):

    __abstract__ = True
    __table_args_ = {
        'mysql_engine': 'InnoDB',
        'mysql_charset': 'utf8'
    }


def reference_col(tablename, nullable=False, pk_name='id', **kwargs):

    return db.Column(
        db.ForeignKey('{0}.{1}'.format(tablename, pk_name)),
        nullable=nullable, **kwargs)


def get_model_registry():
    return (model for model in getattr(db.Model, '_decl_class_registry').values()
            if hasattr(model, '__tablename__'))


def get_model_by_tablename(tablename):
    model_registry = get_model_registry()

    for c in model_registry:
        if hasattr(c, '__tablename__') and c.__tablename__ == tablename:
            return c


def query_result2dict(result_item):
    if callable(getattr(result_item, '_as_dict', None)):
        return getattr(result_item, '_as_dict')()
    elif callable(getattr(result_item, 'to_dict', None)):
        return getattr(result_item, 'to_dict')()
    elif getattr(result_item, '_fields', []):
        return {field: getattr(result_item, field, None)
                for field in getattr(result_item, '_fields')}

    raise ValueError(u'convert query result item {0}, {1} to dict failed'.
                     format(type(result_item), result_item))


def include_custom_types(obj):
    for key in custom_types.__all__:
        if not hasattr(obj, key):
            setattr(obj, key, getattr(custom_types, key))


include_custom_types(db)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/16 by allen
"""

from redis import Redis, ConnectionPool


class RedisManager(Redis):

    def __init__(self, *args, **kwargs):
        super(RedisManager, self).__init__(*args, **kwargs)
        self.connection_conf = {
            'host': 'localhost',
            'port': 6379
        }

        self.connection_pool = None

    def init_app(self, app):
        self.connection_conf = app.config.get(
            'REDIS_CONFIG', self.connection_conf)
        self.connection_pool = ConnectionPool(**self.connection_conf)


extension = RedisManager()



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/25 by allen
"""

import os
import codecs
from setuptools import setup, find_packages

from tadpole.config import VERSION
with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    long_description = f.read()


NAME = "tadpole"
__version__ = VERSION
__author__ = "allen"
__license__ = "MIT"
__copyright__ = "2017 - allen"


description = "flask starter, support gunicorn start,stop,reload script and " \
              "restful view , auto restful sqlalchemy model, auth and  permission"


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


def required_packages():
    with codecs.open("requirements.txt", "r") as pfp:
        return pfp.readlines()


extra_files = package_files('tadpole/template/')
depend_packages = list(required_packages())

setup(
    name=NAME,
    version=__version__,
    license=__license__,
    author=__author__,
    author_email='echoyuanliang@gmail.com',
    description=description,
    long_description=long_description,
    url='https://github.com/echoyuanliang/tadpole',
    download_url='https://github.com/echoyuanliang/tadpole.git',
    install_requires=depend_packages,
    packages=find_packages(),
    package_data={'': extra_files},
    keywords=['flask', 'restful', 'orm', 'sqlalchemy', 'auth', 'permission'],
    platforms='any',
    classifiers=[
        'Operating System :: POSIX',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    entry_points="""
    [console_scripts]
    tadpole=tadpole.run:cli
    """
)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/19 by allen
"""

from datetime import datetime
from app.lib.database import db, Column

user_role = db.Table('user_role',
                     Column('user_id', db.Integer, db.ForeignKey('user.id')),
                     Column('role_id', db.Integer, db.ForeignKey('role.id')),
                     Column('create_date', db.DateTime, nullable=False, default=datetime.now))


role_resource = db.Table('role_resource',
                         Column(
                             'role_id',
                             db.Integer,
                             db.ForeignKey('role.id')),
                         Column(
                             'resource_id',
                             db.Integer,
                             db.ForeignKey('resource.id')),
                         Column('create_date', db.DateTime, nullable=False, default=datetime.now))



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from main import app


@app.rest_route("/health", methods=['GET', "HEAD"])
def health_check(data):
    return dict(code=200, message="I'm OK!!!", data=data)



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/7 by allen
"""

from app.lib.database import db, Column
from app.lib.database import MysqlModel as Model


class User(Model):

    __hide__ = ('password',)

    account = Column(
        db.String(128),
        nullable=False,
        default='-',
        index=True,
        unique=True)
    name = Column(db.String(32), nullable=False, default='-')
    email = Column(db.Email(128), nullable=False, default='-')
    password = Column(db.Password(schemes=['pbkdf2_sha512', 'md5_crypt'],
                                  deprecated=['md5_crypt']), nullable=False, default='-')

    def validate_password(self, password):
        return self.password == password



#!/usr/bin/env python
# coding: utf-8

"""
    create at 2017/11/18 by allen
"""

from functools import partial
from flask import Flask, Blueprint


class PineFlask(Flask):
    from app.lib.utils import CustomJsonEncoder

    json_encoder = CustomJsonEncoder

    def __init__(self, *args, **kwargs):
        from app.lib.rest_route import rest_route

        super(PineFlask, self).__init__(*args, **kwargs)
        self.rest_route = partial(rest_route, self)


class PineBlueprint(Blueprint):

    def __init__(self, *args, **kwargs):
        from app.lib.rest_route import rest_route

        super(PineBlueprint, self).__init__(*args, **kwargs)
        self.rest_route = partial(rest_route, self)
